# -*- cperl -*-
package IndexFile;
use strict;
use warnings;

# インディクスファイルは、各ログディレクトリ内に___index___.txtという名で作られる。
# 中身は次の通り。
#
# DATE=YYYY.MM.DD; SIZE=1830; PRIVMSG-COUNT=294; NOTICE-COUNT=185; COUNT-00=10; COUNT-01=11; ...
# DATE=YYYY.MM.DD; SIZE=3810; PRIVMSG-COUNT=632; NOTICE-COUNT=39; COUNT-00=53; COUNT-01=102; ...
# ...

sub get {
    # $logdir: ログディレクトリ名 (例: logs/NanminGoya/200401)
    # $day   : 日付 (例: 2004.01.01)
    my ($logdir, $day) = @_;
    my $index_fpath = "$logdir/___index___.txt";
    
    if (!-r $index_fpath) {
	die "インディクスファイルが存在しないか、パーミッションを落とされています。\n".
	  "ファイル名: $index_fpath\n".
	  "もしこのエラーが1時間以上に渡って表示され続けるようでしたら、お手数ですがサイト管理者へご連絡下さい。";
    }

    foreach my $index (_load($index_fpath)) {
	if ($index->{DATE} eq $day) {
	    return $index;
	}
    }
    undef;
}

sub update {
    # $logdir: ログディレクトリ名 (例: logs/NanminGoya/200401)
    my $logdir = shift;
    my $index_fpath = "$logdir/___index___.txt";

    my @index;
    if (-r $index_fpath) {
	@index = _load($index_fpath);
    }

    # ファイル一覧を得る
    my @files = do {
	# YYYY.MM.DDの配列
	opendir my $dh, $logdir
	  or die "Failed to opendir $logdir";
	my @dates;
	while (defined($_ = readdir $dh)) {
	    -f "$logdir/$_" and m/^(\d{4}\.\d{2}\.\d{2})\.txt$/ and
	      push @dates, $1;
	}
	@dates;
    };

    my @del;
    my %dates_hash;
    my $is_dirty;
    foreach my $item (@index) {
	my $logfile_fpath = "$logdir/$item->{DATE}.txt";
	if (!-f $logfile_fpath) {
	    # 消えてる。
	    push @del, $item;
	}
	else {
	    $dates_hash{$item->{DATE}} = 1;
	    
	    my $size = (stat($logfile_fpath))[7];
	    if ($item->{SIZE} != $size) {
		# 更新されてる。
		my $stat = _make_stat($logfile_fpath);
		%$item = %$stat;
		$is_dirty = 1;
	    }
	}
    }

    foreach my $item (@del) {
	@index = grep { $_ != $item } @index;
	$is_dirty = 1;
    }
    
    foreach (@files) {
	$dates_hash{$_} and next;

	# 追加されてる
	push @index, _make_stat("$logdir/$_.txt");

	$is_dirty = 1;
    }

    if ($is_dirty) {
	_store(\@index, $index_fpath);
    }
}

sub _make_stat {
    my $fpath = shift;
    my $stat = {};

    if (!-f $fpath) {
	die "_make_stat: !-f $fpath";
    }
    
    if ($fpath =~ m,(?:^|/)(\d{4}\.\d{2}\.\d{2})\.txt$,) {
	$stat->{DATE} = $1;
    }
    else {
	die "_make_stat: $fpath is not an irc log file";
    }

    $stat->{SIZE} = (stat($fpath))[7];

    # PRIVMSGのフォーマット:
    # HH:MM:SS <&難民小屋> ぽに

    # NOTICEのフォーマット:
    # HH:MM:SS (&難民小屋) |゜Д）ぽに

    open my $fh, '<', $fpath
      or die "_make_stat: Failed to read file $fpath";

    # 時間帯別の発言回数をカウントする。PRIVMSGとNOTICEは合算。
    $stat->{'PRIVMSG-COUNT'} = 0;
    $stat->{'NOTICE-COUNT'} = 0;
    foreach (0 .. 23) {
	$stat->{sprintf 'COUNT-%02d', $_} = 0;
    }
    
    my $add = sub {
	my $hour = shift;
	$stat->{sprintf 'COUNT-%02d', $hour}++;
    };
    
    while (<$fh>) {
	if (/^(\d\d):\d\d(?::\d\d)?\s+<.+?>\s+/) {
	    $stat->{'PRIVMSG-COUNT'}++;
	    $add->($1);
	}
	elsif (/^(\d\d):\d\d(?::\d\d)?\s+\(.+?\)\s+/) {
	    $stat->{'NOTICE-COUNT'}++;
	    $add->($1);
	}
    }
    
    $stat;
}

my $INDEX_CACHE;
sub _load {
    my $fpath = shift;

    $INDEX_CACHE ||= {};
    
    if ($_ = $INDEX_CACHE->{$fpath}) {
	return @$_;
    }
    
    open my $fh, '<', $fpath
      or die "Failed to read file $fpath";
    flock $fh, 1; # LOCK_SH

    my @index;
    while (<$fh>) {
	tr/\r\n//d;
	m/^#/ and next;
	s/^\s*|\s*$//g;
	length or next;

	push @index, _makedict($_);
    }
    
    flock $fh, 4; # LOCK_UN
    close $fh;

    $INDEX_CACHE->{$fpath} = \@index;

    @index;
}

sub _makedict {
    my $str = shift;
    my %dict;

    foreach (split m/\s*;\s*/, $str) {
	/^(.+?)\s*=\s*(.+)$/ or
	  die "Bad dict in the conf: $_";

	$dict{$1} = $2;
    }
    \%dict;
}

sub _strdict {
    my $dict = shift;
    join(
	'; ',
	map {
	    "$_=$dict->{$_}";
	} keys %$dict);
}

sub _store {
    my ($index, $fpath) = @_;
    open my $fh, '>>', $fpath
      or die "Failed to write file $fpath";
    flock $fh, 2; # LOCK_EX
    truncate $fh, 0;
    seek $fh, 0, 0;

    print $fh "# -*- fundamental -*-\n";
    print $fh "# This is an auto-generated index file for the irclog cgi.\n";
    print $fh "# DO NOT EDIT THIS.\n";
    print $fh "\n";

    foreach (@$index) {
	print $fh _strdict($_), "\n";
    }

    print $fh "\n";
    print $fh "# End of file\n";

    flock $fh, 4; # LOCK_UN
    close $fh;
}

1;
